import * as Core from '@ts-foundation/core';
import { Checkbox } from '@ts-foundation/core';

type Undefinable<T extends object> = {
    [K in keyof T]: T[K] extends undefined ? T[K] : T[K] | undefined;
};
type Projection<T extends object> = Core.Combine<{
    [K in (keyof T & (string | number)) as T[K] extends Function ? never : K]: T[K];
}>;
type Select<TSelect extends object, TFrom extends {
    [k in keyof TSelect]: any;
}> = Projection<Core.Select<TSelect, TFrom>>;
type Subtract<TSubtract extends object, TFrom extends {
    [k in keyof TSubtract]: any;
}> = Projection<Core.Subtract<TSubtract, TFrom>>;
type Difference<T1 extends object, T2 extends object> = Projection<Core.Difference<T1, T2>>;
type Intersection<T1 extends object, T2 extends object> = Projection<Core.Intersection<T1, T2>>;
type Merge<T1 extends object, T2 extends object> = Projection<Core.Merge<T1, T2>>;
type Override<TFields extends object, TFrom extends {
    [K in keyof TFields]: unknown;
}> = Projection<Core.Override<TFields, TFrom>>;
type Extend<TFields extends object, TFrom extends object> = Projection<Core.Extend<TFields, TFrom>>;
type DeepProjection<T extends object> = Core.Combine<{
    [K in (keyof T & (string | number)) as T[K] extends Function ? never : K]: T[K] extends object ? T[K] extends Array<infer U> ? U extends object ? DeepProjection<U>[] : U[] : DeepProjection<T[K]> : T[K];
}>;
type Immutable<T extends object> = Core.Combine<Readonly<{
    [K in (keyof T & (string | number)) as T[K] extends Function ? never : K]: T[K];
}>>;
type DeepImmutable<T extends object> = Core.Combine<Readonly<{
    [K in (keyof T & (string | number)) as T[K] extends Function ? never : K]: T[K] extends object ? T[K] extends Array<infer U> ? U extends object ? DeepImmutable<U>[] : U[] : DeepImmutable<T[K]> : T[K];
}>>;
type SwitchSchema<T extends object, TSchema extends Record<(keyof T | string), string>> = Projection<{
    -readonly [K in keyof TSchema as TSchema[K]]: K extends keyof T ? T[K] : never;
}>;
type FlipSchema<T extends Record<string, string>> = Readonly<Projection<{
    [K in keyof T as T[K]]: K;
}>>;
type Lens<T extends object> = Projection<Record<Core.ObjectKeys<T>, undefined>>;
type LensProjection<TData extends object, TLens extends Record<string, undefined>> = Undefinable<Projection<{
    [K in keyof TLens]: K extends keyof TData ? TData[K] : never;
}>>;
type GroupBy<T extends object, K extends string = string> = Record<K, T[]>;

declare const difference: <T1 extends object, T2 extends object>(a: T1, b: T2) => Difference<T1, T2>;
declare const intersection: <T1 extends object, T2 extends object>(a: T1, b: T2) => Intersection<T1, T2>;
declare const merge: <T1 extends object, T2 extends object>(a: T1, b: T2) => Merge<T1, T2>;
declare const switchSchema: <T extends object, TSchema extends Record<(keyof T | string), string>>(args: T, schema: TSchema) => SwitchSchema<T, TSchema>;
declare const flipSchema: <TSchema extends Record<string, string>>(schema: TSchema) => FlipSchema<TSchema>;
declare const immutable: <TData extends object>(data: TData) => DeepImmutable<TData>;
declare const projection: <TData extends object>(data: TData) => DeepProjection<TData>;
declare const lens: <TIn extends object, TLens extends Lens<Checkbox<TIn>> = { [K_4 in (Checkbox<TIn> extends infer T_1 ? { [K_1 in keyof T_1]: Checkbox<TIn>[K_1] extends Function ? never : K_1; } : never)[Exclude<keyof TIn, keyof TIn> | keyof (keyof TIn extends infer T_2 extends keyof TIn ? { [K_2 in T_2]-?: Required<Pick<TIn, K_2>> & Partial<Pick<TIn, Exclude<keyof TIn, K_2>>> extends infer T_3 ? { [K_3 in keyof T_3]: (Required<Pick<TIn, K_2>> & Partial<Pick<TIn, Exclude<keyof TIn, K_2>>>)[K_3]; } : never; } : never)[keyof TIn]] & string as Record<Core.ObjectKeys<Checkbox<TIn>>, undefined>[K_4] extends Function ? never : K_4]: Record<Core.ObjectKeys<Checkbox<TIn>>, undefined>[K_4]; } extends infer T ? { [K in keyof T]: { [K_4 in (Checkbox<TIn> extends infer T_1 ? { [K_1 in keyof T_1]: Checkbox<TIn>[K_1] extends Function ? never : K_1; } : never)[Exclude<keyof TIn, keyof TIn> | keyof (keyof TIn extends infer T_2 extends keyof TIn ? { [K_2 in T_2]-?: Required<Pick<TIn, K_2>> & Partial<Pick<TIn, Exclude<keyof TIn, K_2>>> extends infer T_3 ? { [K_3 in keyof T_3]: (Required<Pick<TIn, K_2>> & Partial<Pick<TIn, Exclude<keyof TIn, K_2>>>)[K_3]; } : never; } : never)[keyof TIn]] & string as Record<Core.ObjectKeys<Checkbox<TIn>>, undefined>[K_4] extends Function ? never : K_4]: Record<Core.ObjectKeys<Checkbox<TIn>>, undefined>[K_4]; }[K]; } : never, TData extends Partial<TIn> = Partial<TIn>>(data: TData, from: TLens) => LensProjection<TData, TLens>;
declare const groupBy: <T extends object, K extends string = string>(entries: T[], by: (entry: T) => K) => GroupBy<T, K>;

declare const projections_difference: typeof difference;
declare const projections_flipSchema: typeof flipSchema;
declare const projections_groupBy: typeof groupBy;
declare const projections_immutable: typeof immutable;
declare const projections_intersection: typeof intersection;
declare const projections_lens: typeof lens;
declare const projections_merge: typeof merge;
declare const projections_projection: typeof projection;
declare const projections_switchSchema: typeof switchSchema;
declare namespace projections {
  export { projections_difference as difference, projections_flipSchema as flipSchema, projections_groupBy as groupBy, projections_immutable as immutable, projections_intersection as intersection, projections_lens as lens, projections_merge as merge, projections_projection as projection, projections_switchSchema as switchSchema };
}

export { type DeepImmutable, type DeepProjection, type Difference, type Extend, type FlipSchema, type GroupBy, type Immutable, type Intersection, type Lens, type LensProjection, type Merge, type Override, type Projection, projections as Projections, type Select, type Subtract, type SwitchSchema, type Undefinable, difference, flipSchema, groupBy, immutable, intersection, lens, merge, projection, switchSchema };
