var __defProp = Object.defineProperty;
var __export = (target, all) => {
  for (var name in all)
    __defProp(target, name, { get: all[name], enumerable: true });
};

// src/libs/projections.ts
var projections_exports = {};
__export(projections_exports, {
  difference: () => difference,
  flipSchema: () => flipSchema,
  groupBy: () => groupBy,
  immutable: () => immutable,
  intersection: () => intersection,
  lens: () => lens,
  merge: () => merge,
  projection: () => projection,
  switchSchema: () => switchSchema
});
var difference = (a, b) => {
  const uniqueKeys = /* @__PURE__ */ new Set([
    ...Object.keys(a).filter(
      (k) => !(k in b) && typeof a[k] !== "function"
    ),
    ...Object.keys(b).filter(
      (k) => !(k in a) && typeof b[k] !== "function"
    )
  ]);
  const entries = Array.from(uniqueKeys).sort().map((key) => [
    key,
    key in a ? a[key] : b[key]
  ]);
  return Object.fromEntries(entries);
};
var intersection = (a, b) => {
  const entries = Object.keys(a).filter(
    (k) => k in b && typeof a[k] !== "function" && typeof b[k] !== "function"
  ).sort().map((key) => [key, b[key]]);
  return Object.fromEntries(entries);
};
var merge = (a, b) => {
  const uniqueKeys = /* @__PURE__ */ new Set([
    ...Object.keys(a).filter(
      (k) => typeof a[k] !== "function"
    ),
    ...Object.keys(b).filter(
      (k) => typeof b[k] !== "function"
    )
  ]);
  const entries = Array.from(uniqueKeys).sort().map((key) => [
    key,
    key in b ? b[key] : a[key]
  ]);
  return Object.fromEntries(entries);
};
var switchSchema = (args, schema) => {
  const entries = Object.entries(args).filter(
    ([key, value]) => key in schema && typeof value !== "function"
  ).map(([key, value]) => [
    schema[key],
    value
  ]);
  return Object.fromEntries(entries);
};
var flipSchema = (schema) => {
  const entries = Object.entries(schema).map(([key, value]) => [value, key]);
  return Object.fromEntries(entries);
};
var immutable = (data) => {
  const entries = Object.entries(data).filter(([, value]) => typeof value !== "function").map(([key, value]) => [
    key,
    value && typeof value === "object" ? immutable(value) : value
  ]);
  return Object.freeze(
    Object.fromEntries(entries)
  );
};
var projection = (data) => {
  const entries = Object.entries(data).filter(([, value]) => typeof value !== "function").sort(([a], [b]) => a.localeCompare(b)).map(([key, value]) => [
    key,
    value && typeof value === "object" ? Array.isArray(value) ? value : projection(value) : value
  ]);
  return Object.fromEntries(entries);
};
var lens = (data, from) => {
  const entries = Object.entries(from).filter(
    ([key]) => key in data && typeof data[key] !== "function"
  ).map(([key]) => [key, data[key]]);
  return Object.fromEntries(entries);
};
var groupBy = (entries, by) => {
  const result = {};
  for (const entry of entries) {
    const key = by(entry);
    if (!result[key]) {
      result[key] = [];
    }
    result[key].push(entry);
  }
  return result;
};
export {
  projections_exports as Projections,
  difference,
  flipSchema,
  groupBy,
  immutable,
  intersection,
  lens,
  merge,
  projection,
  switchSchema
};
